import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;

import com.sun.java.swing.*;
import com.sun.java.swing.preview.*;
import com.sun.java.swing.event.*;

import sv.kernel.*;
import sv.util.*;
import sv.ui.*;

/** 
 * GFrame2D.java
 * Define a frame to draw the 2D graph
 * @version 1.0    10/15/97
 * @author Byeongseob Ki
 */

public class GFrame2D extends GFrame implements WindowListener {

  public static final int WIREMODE       = 0;
  public static final int WIRECOLORMODE  = 1;
  public static final int HIDDENMODE     = 2;
  public static final int SPECTRUMMODE   = 3;
  public static final int GRAYMODE       = 4;
  public static final int SURFACEMODE    = 5;
  public static final int CONTOURMODE    = 6;
  public static final int POINTMODE      = 7;

  svserver svServer;

  CutCopyManager editManager;

  // animation class
  private Animator animator; 

  // animation speed
  private int animationSpeed = 300;

  // Thread to creat GIF image
  private GIFer gifer;

  /**
   * to access GCanvas2D class
   */
  GCanvas2D canvas;

  // various MenuItems
  private MenuItem animateMI, stopMI, nextMI, prevMI, splitMI;
  public CheckboxMenuItem boundMI, meshMI, faceMI, xytickMI, ztickMI,
                  surfaceMI, wireMI, cwireMI, hiddenMI, spectrumMI, 
                  grayMI, contourMI, vectorMI, stealthMI, overlayMI,
                  timeMI, delayMI, globalMI, pointMI, discardMI,
                  xscaleMI, yscaleMI, zscaleMI, allscaleMI;        



  // global min and max value
  public float globalxmin = Float.MAX_VALUE, globalxmax = Float.MIN_VALUE;
  public float globalymin = Float.MAX_VALUE, globalymax = Float.MIN_VALUE;
  public float globalzmin = Float.MAX_VALUE, globalzmax = Float.MIN_VALUE;

  // contour setting 
  int contour_lines = 5;
  float contour_min;
  float contour_max;

  int mergeIndex = 1;
  boolean mergeSelected = false;
  boolean equallyspaced = true;
                             
  // configuration
  public Config2D config;

  JSlider zoomer;
  Label zoomValue;
  int zfactor = 20;
  int zmethod = 0;    // 0 = all, 1 = x scale,  2 =  y scale, 3 = z scale

  /** 
   * constructor
   * @param title title of this window
   * @param dataSet 2D data set
   * @param sv  svserver
   */
  public GFrame2D(String title, DataSet2D dataSet, 
                                   Config2D config, svserver svServer) {
     super(title);

     this.svServer = svServer;
     dimension_ = DataSet.TWOD;
     this.config = config;

     // copy the DataSet
     dataSetVector.addElement(dataSet);

     noOfTime = dataSet.getnoOfTime();
     timeArr = dataSet.getTimeArr();

     for (int i=0; i<noOfTime; i++) {

         if (globalxmin > dataSet.timeData[i].getxmin()) 
            globalxmin = dataSet.timeData[i].getxmin();
         if (globalxmax < dataSet.timeData[i].getxmax()) 
            globalxmax = dataSet.timeData[i].getxmax();

         if (globalymin > dataSet.timeData[i].getymin()) 
            globalymin = dataSet.timeData[i].getymin();
         if (globalymax < dataSet.timeData[i].getymax()) 
            globalymax = dataSet.timeData[i].getymax();

         if (globalzmin > dataSet.timeData[i].getzmin()) 
            globalzmin = dataSet.timeData[i].getzmin();
         if (globalzmax < dataSet.timeData[i].getzmax()) 
            globalzmax = dataSet.timeData[i].getzmax();
      }

      makeMenu();

      boolean b = false;

      if (config.colormap_2d != null) {
         b = readColorMap(config.colormap_2d);

         if (b) {
           float value;

           for (int i=0; i<101; i++) {
              value = rspline.computePoint((float)i/100f); 
              if (value > 1.0f) value = 1.0f;
              else if (value < 0.0f) value = 0.0f;
              rcolors[i] = value;

              value = gspline.computePoint((float)i/100f);    
              if (value > 1.0f) value = 1.0f;
              else if (value < 0.0f) value = 0.0f;
              gcolors[i] = value;

              value = bspline.computePoint((float)i/100f);    
              if (value > 1.0f) value = 1.0f;
              else if (value < 0.0f) value = 0.0f;
              bcolors[i] = value;
           }
         }
     }

     if (config.colormap_2d == null || b == false) {
         int nControls = 5;
         float[] xcontrols = { 0.0f, 0.25f, 0.5f, 0.75f, 1.0f };
         float[] ycontrols = { 0.0f, 0.5f, 1.0f, 0.5f, 0.0f };
         rspline = new NewtonMethod(xcontrols, ycontrols);
         hspline = new NewtonMethod(xcontrols, ycontrols);

         for (int i=0; i<5; i++)
           ycontrols[i] = 0.5f;
         gspline = new NewtonMethod(xcontrols, ycontrols);
         sspline = new NewtonMethod(xcontrols, ycontrols);

         ycontrols[0] = 1.0f;
         ycontrols[1] = 0.5f; 
         ycontrols[2] = 0.0f;
         ycontrols[3] = 0.5f;
         ycontrols[4] = 1.0f;
         bspline = new NewtonMethod(xcontrols, ycontrols);
         lspline = new NewtonMethod(xcontrols, ycontrols);
     }

     setLayout(new BorderLayout());
     canvas = new GCanvas2D(this);

     add("Center", canvas);
     Image img = svServer.loadImageIcon(getClass().getResource("icons/wheel.gif"), "wheel").getImage();

     add("East", new Wheel(Wheel.VERTICAL, 140, canvas, img));

     Panel p2 = new Panel();
     p2.setLayout(new BorderLayout());

     JPanel p = new JPanel();
     p.setLayout(new BoxLayout(p, BoxLayout.X_AXIS));
     p.add(Box.createRigidArea(new Dimension(20, 1)));
     p.add(new Wheel(Wheel.HORIZONTAL, 140, canvas, img));
     p.add(Box.createRigidArea(new Dimension(20, 1)));

     Panel zp = new Panel();
     zp.setLayout(new BorderLayout());
     zp.add("West", new Label("Zoom:"));
     zp.add("Center", zoomer = new JSlider(JSlider.HORIZONTAL, 0, 60, zfactor));   
     zp.add("East",  zoomValue = new Label("1.0"));

     p.add(zp);
     p.add(Box.createRigidArea(new Dimension(20, 1)));
     createSlideListener();

     p2.add("North", p);
     p2.add("Center", monitor = new SimpleMonitor());
     add("South", p2);
     
     monitor.display(title + " created.", false);

     calcVertex();

     pack();
     validate();
     addWindowListener(this); 
  }

  /**
   * update filter menu
   */
  public void updateFilterMenu() {
     filterMenu.removeAll();
     createFilterMenu();
  }


  /**
   * add another 2D data set to this frame
   * @param dataSet DataSet to be added
   */
  public void addDataSet(DataSet2D dataSet) {
     // reset all the currentTime value to 0
     currentTime = 0;

     dataSetVector.addElement(dataSet);
     updateTimeArr();
     splitMI.setEnabled(true);
     calcVertex();

     monitor.display("Added new DataSet " + dataSet.getTitle(), false);
     setFrameTitle(dataSet.getTitle() + "...");
     canvas.makeContourCache();
     canvas.repaint();
  }


  private void calcGlobalMinMax() {
     globalxmin = Float.MAX_VALUE; globalxmax = Float.MIN_VALUE;
     globalymin = Float.MAX_VALUE; globalymax = Float.MIN_VALUE;
     globalzmin = Float.MAX_VALUE; globalzmax = Float.MIN_VALUE;

     for (int j=0; j<dataSetVector.size(); j++) {
        DataSet2D data = (DataSet2D)dataSetVector.elementAt(j);
        for (int i=0; i<data.getnoOfTime(); i++) {

           if (globalxmin > data.timeData[i].getxmin()) 
              globalxmin = data.timeData[i].getxmin();
           if (globalxmax < data.timeData[i].getxmax()) 
              globalxmax = data.timeData[i].getxmax();

           if (globalymin > data.timeData[i].getymin()) 
              globalymin = data.timeData[i].getymin();
           if (globalymax < data.timeData[i].getymax()) 
              globalymax = data.timeData[i].getymax();

           if (globalzmin > data.timeData[i].getzmin()) 
              globalzmin = data.timeData[i].getzmin();
           if (globalzmax < data.timeData[i].getzmax()) 
              globalzmax = data.timeData[i].getzmax();
        }
     }
  }


  public synchronized void calcVertex() {

    if (globalFlag) { 
       calcGlobalMinMax();
       canvas.setRanges(globalxmin, globalxmax, globalymin, globalymax);
    } else {
       float xxmin = 1e10f, xxmax = 1e-10f, 
             yymin = 1e10f, yymax = 1e-10f;

       for (Enumeration e=dataSetVector.elements(); e.hasMoreElements();) {
          DataSet2D data = (DataSet2D)e.nextElement();

          for (int i=0; i<data.getnoOfTime(); i++) {
            if (data.getTimeValue(i) == timeArr[currentTime]) {
              if (xxmin > data.timeData[i].getxmin())
               xxmin = data.timeData[i].getxmin();
              if (xxmax < data.timeData[i].getxmax())
               xxmax = data.timeData[i].getxmax();
              if (yymin > data.timeData[i].getymin())
               yymin = data.timeData[i].getymin();
              if (yymax < data.timeData[i].getymax())
               yymax = data.timeData[i].getymax();
              break;
	    }
          }
       }
       canvas.setRanges(xxmin, xxmax, yymin, yymax);
    }
  
    canvas.destroyImage();
  
    float[] minmaxArr = new float[4];
    minmaxArr[0] = globalxmin;
    minmaxArr[1] = globalxmax;
    minmaxArr[2] = globalymin;
    minmaxArr[3] = globalymax;

    int index = 0;
    for (Enumeration e=dataSetVector.elements(); e.hasMoreElements();) {
      DataSet2D data = (DataSet2D)e.nextElement();
      for (int i=0; i<data.getnoOfTime(); i++) {
         data.timeData[i].makeVertex(globalFlag, minmaxArr);
         data.timeData[i].normalizeVectorSet(globalFlag, minmaxArr);
      }
      index++;
    }

    equallyspaced = isEquallySpaced();
    canvas.setDataAvailability(true);
    canvas.repaint(); 
  }


  /**
   * normaize VectorSet
   */
  public void normalizeVectorSet() {
  
    float[] minmaxArr = new float[4];
    minmaxArr[0] = globalxmin;
    minmaxArr[1] = globalxmax;
    minmaxArr[2] = globalymin;
    minmaxArr[3] = globalymax;

    for (Enumeration e=dataSetVector.elements(); e.hasMoreElements();) {
      DataSet2D data = (DataSet2D)e.nextElement();
      for (int i=0; i<data.getnoOfTime(); i++) 
       data.timeData[i].normalizeVectorSet(globalFlag, minmaxArr);
    }
  }

  /**
   * Determines whether the delay regeneration checkbox is checked.
   */
  public boolean isExpectDelay() {
    return delayMI.getState();
  }

  /**
   * Determines whether to show bounding box.
   */
  public boolean isBoxed() {
    return boundMI.getState();
  }

  /**
   * Determines whether to show x-y mesh.
   */
  public boolean isMesh() {
    return meshMI.getState();
  }


  /**
   * Determines whether to show x-y ticks.
   */
  public boolean isDisplayXY() {
    return xytickMI.getState();
  }

  /**
   * Determines whether to show z ticks.
   */
  public boolean isDisplayZ() {
    return ztickMI.getState();
  }
  
  /**
   * Determines whether to show face grids.
   */
  public boolean isDisplayGrids() {
    return faceMI.getState();
  }


  /**
   * Determines whether to show vector plot
   */
  public boolean isVectorPlot() {
    return vectorMI.getState();
  }

  /**
   * Gets the selected shading/plot mode
   * @return the selected shading/plot mode
   */
   
  public int getPlotMode() {    
    if (wireMI.getState()) return WIREMODE;
    else if (cwireMI.getState()) return WIRECOLORMODE;
    else if (hiddenMI.getState()) return HIDDENMODE;
    else if (spectrumMI.getState()) return SPECTRUMMODE;
    else if (grayMI.getState()) return GRAYMODE;
    else if (pointMI.getState()) return POINTMODE;
    return WIREMODE;
  }
  
  /**
   * check if data sets are equally spaced
   */
  private boolean isEquallySpaced() {
    boolean b = true;

    for (Enumeration e=dataSetVector.elements(); e.hasMoreElements();) {
       DataSet2D data = (DataSet2D)e.nextElement();

       for (int i=0; i<data.getnoOfTime(); i++) {
	 if (!data.timeData[i].isEquallySpaced()) 
             return false;
       }
    }
    return b;
  }

  /**
   * Gets minimum z
   */
  public float getZMin() {
    if (globalFlag) {
       contour_min = globalzmin;
       return globalzmin;
    } else {
      float zzmin = 1e10f;
      for (Enumeration e=dataSetVector.elements(); e.hasMoreElements();) {
         DataSet2D data = (DataSet2D)e.nextElement();

         for (int i=0; i<data.getnoOfTime(); i++) {
            if (data.getTimeValue(i) == timeArr[currentTime] &&
		zzmin > data.timeData[i].getzmin()) {
               zzmin = data.timeData[i].getzmin();
               break;
            }
         }
      }
      contour_min = zzmin;
      return zzmin;
    }
  }

  /**
   * Gets maximum z 
   */
  public float getZMax() {
    if (globalFlag) {
       contour_max =  globalzmax;
       return globalzmax;
    } else {
      float zzmax = -1e10f;
      for (Enumeration e=dataSetVector.elements(); e.hasMoreElements();) {
         DataSet2D data = (DataSet2D)e.nextElement();

         for (int i=0; i<data.getnoOfTime(); i++) {
            if (data.getTimeValue(i) == timeArr[currentTime] &&
		zzmax < data.timeData[i].getzmax()) {
               zzmax = data.timeData[i].getzmax();
               break;
            }
         }
      }
      contour_max = zzmax;
      return zzmax;
    }
  }


  /**
   * Gets minimum c
   */
  public float getCMin() {

    float ccmin = 1e10f;
    for (Enumeration e=dataSetVector.elements(); e.hasMoreElements();) {
       DataSet2D data = (DataSet2D)e.nextElement();

       for (int i=0; i<data.getnoOfTime(); i++) {
          if (data.getTimeValue(i) == timeArr[currentTime] &&
		ccmin > data.timeData[i].getcmin()) {
             ccmin = data.timeData[i].getcmin();
             break;
          }
       }
    }
    return ccmin;

  }

  /**
   * Gets maximum c 
   */
  public float getCMax() {

    float ccmax = -1e10f;
    for (Enumeration e=dataSetVector.elements(); e.hasMoreElements();) {
       DataSet2D data = (DataSet2D)e.nextElement();

       for (int i=0; i<data.getnoOfTime(); i++) {
          if (data.getTimeValue(i) == timeArr[currentTime] &&
		ccmax < data.timeData[i].getcmax()) {
             ccmax = data.timeData[i].getcmax();
             break;
          }
       }
    }
    return ccmax;

  }


  /**
   * set color map
   */
  public void setColorMap(String[] str) {
     super.setColorMap(str);
     canvas.destroyImage();
     canvas.repaint();
  }

  
  /**
   * Gets the number of contour lines to be created
   * @return the number of contour lines
   */
  public int getContourLines() {
    return contour_lines;
  }
  

  /**
   * Sets the number of contour lines to be created
   * @param n the number of contour lines
   */
  public void setContourLines(int n) {
    contour_lines = n;
  }

  /**
   * Sets contour min value 
   * @param val contour min value 
   */
  public void setContourMin(float val) {
    contour_min = val;
  }  
 

  /**
   * Sets contour max value 
   * @param val contour max value 
   */
  public void setContourMax(float val) {
    contour_max = val;
  }  

  /**
   * Gets minimum contour z
   */
  public float getContourMin() {
     return contour_min;
  }

  /**
   * Gets maximum contour z 
   */
  public float getContourMax() {
     return contour_max;
  }


  // type one calulation. (Filter.NORMAL)
  public void doTypeOne(Filter filter, String input, boolean broad) {

     long newFrameID = -1;
     boolean firstTime = true;
     GFrame2D otherFrame2D = null;
     GFPlane otherFrame = null;
     Vector newDatasets = new Vector();

     int noOfInputs = filter.getNoOfArgs();

     int dsize = dataSetVector.size();

     for (int elt=0; elt<dsize; elt++) {
        DataSet2D data = (DataSet2D)dataSetVector.elementAt(elt);
        filter.clear();

        if (noOfInputs == 1) {
           float val = 0.0f;
           try {
              val = Float.valueOf(input).floatValue();
           } catch (Exception ee) {
              monitor.display("Input should be number...", false);
              return;
           }
           
           filter.assignDataSet(data);
           filter.assignArgs(val);

        } else if (noOfInputs == 2) {
           float start = 0, end = 0;
           int where = input.indexOf("-");
           if (where < 0) {
              monitor.display("Usage: start_time - end_time", false);
              return;
           } else {
              try {
                start = 
                 Float.valueOf(input.substring(0, where).trim()).floatValue();
                end = 
                 Float.valueOf(input.substring(where+1).trim()).floatValue();
              } catch (Exception ee) {
                  monitor.display("Usage: start - end", false);
                  return;
              }
  
              if (end < start) {
                  monitor.display("Invalid range ...", true);
                  return;
              }
          }
          filter.assignDataSet(data);
          filter.assignArgs(start, end);

        } else    // if it doesn't need an argument
           filter.assignDataSet(data);

        if (filter.getOutDimension() == 2) {
          DataSet2D newdata;

	  try { 
             newdata = (DataSet2D)filter.performFilter();
	  } catch (Exception ee) {
             monitor.display("Filter can't perform. Check filter codes.", true);
             return;
          }

          if (config.overlay_mode) {
             addDataSet(newdata);
             if (broad)
	        svServer.cmanager.broadcastDataSet(newdata, id);

          } else if (config.discard_mode) {
             newDatasets.addElement(newdata);
             if (broad)
	        svServer.cmanager.broadcastDataSet(newdata, id);
 
          } else {
             if (firstTime) {
                firstTime = false;

                if (newdata == null)
                  return;

                newFrameID = svServer.fmanager.register(newdata, -1, null);
                otherFrame2D = (GFrame2D)svServer.fmanager.getGFrame(newFrameID);
                if (broad) 
                   svServer.cmanager.broadcastDataSet(newdata, newFrameID);
              } else {
                if (newdata == null)
                  return;

                 otherFrame2D.addDataSet(newdata);
                 if (broad) 
                    svServer.cmanager.broadcastDataSet(newdata, newFrameID);
              }
	  }

	} else if (filter.getOutDimension() == 1) {
          DataSet1D newdata;

	  try { 
             newdata = (DataSet1D)filter.performFilter();
	  } catch (Exception ee) {
             monitor.display("Filter can't perform. Check filter codes.", true);
             return;
          }

          if (newdata == null) {
              monitor.display("No matching values ...", true);
              return;
           }
          

           if (firstTime) {
              firstTime = false;

              if (newdata == null)
                return;

              newFrameID = svServer.fmanager.register(newdata, -1, null);
              otherFrame = (GFPlane)svServer.fmanager.getGFrame(newFrameID);
              if (broad) 
                svServer.cmanager.broadcastDataSet(newdata, newFrameID);
            } else {
              if (newdata == null)
                 return;

               otherFrame.addDataSet(newdata);
               if (broad) 
                  svServer.cmanager.broadcastDataSet(newdata, newFrameID);
            }

        } else {
           monitor.display("Not supported yet...", true);
           System.out.println("This kind of filter is not support yet.");
           System.out.println("Please let me know your filter's description");
           System.out.println("to bki@top.cis.syr.edu");
           return;
        }
     }

     if (filter.getOutDimension() == 2 && config.discard_mode) {
         dataSetVector = newDatasets;
         updateTimeArr();
         calcVertex();
         canvas.makeContourCache();
         canvas.destroyImage();
         canvas.repaint();
     }
  }

  /**
   * do filter action
   * @param command action command
   */
  public void doFilterAction(String command) {
    Vector menuVec = svServer.menuHandler.menuVec2D;
    for (Enumeration e=menuVec.elements(); e.hasMoreElements();) {
       MenuInfo info = (MenuInfo)e.nextElement();
       if (command.equals(info.name)) {

          Filter filter = null;
          try {
             String dir = "";
             if (info.directory != null)
                dir = info.directory;

             filter=(Filter)Class.forName(dir+"."+info.className).newInstance();

          } catch (Exception exception) {
             System.out.println("Can't find " + info.className + "...");
             return;
          }

          if (filter.getStyle() == Filter.NORMAL) {

             if (filter.getNoOfArgs() == 0) 
                svServer.fmanager.actionFilter(id, DataSet.TWOD, null,
                                             filter, null, true);
             else {
               String result = 
                    JOptionPane.showInputDialog(this, filter.getLabel());
               if (result == null) return;
               svServer.fmanager.actionFilter(id, DataSet.TWOD, null,
                                             filter, result, true);
             }
             break;
          } else if (filter.getStyle() == Filter.MULTIPLE) {
             svServer.mergeHandler.mergeMode2D = true;
             filter.clear();
             svServer.mergeHandler.mergeFilter2D = filter;
             mergeIndex = svServer.mergeHandler.mergeIndex2D = 1;
             mergeSelected = true;
             svServer.mergeHandler.mergeIndex2D++;
             canvas.destroyImage();
             canvas.repaint();

             filter.assignDataSet(new Long(id));

             if (filter.getExpectDataSets() < 2) {
               monitor.display("Check number of frames to be merged...", true);
               return;
             }
             monitor.display("Click the mouse on second GraphFrame ...", 
                                                           false);
	  }
       }
     }
  }

  /**
   * internal merge action
   * @param b if control key is pressed
   */
  public void mergeAction(boolean b) {

      int howmany = svServer.mergeHandler.mergeFilter2D.getDataVectorSize();
      mergeIndex = svServer.mergeHandler.mergeIndex2D;
      mergeSelected = true;
      svServer.mergeHandler.mergeIndex2D++;
      canvas.destroyImage();
      canvas.repaint();
      
      if (svServer.mergeHandler.mergeFilter2D.getExpectDataSets() > 100) {
	if (b) {
          svServer.mergeHandler.mergeMode2D = false;
          svServer.mergeHandler.mergeFilter2D.assignDataSet(new Long(id));
          svServer.mergeHandler.mergeFrames2D();
        } else {
          svServer.mergeHandler.mergeFilter2D.assignDataSet(new Long(id));
          monitor.display("merged ...", false);
        }
      } else if (howmany == 
              (svServer.mergeHandler.mergeFilter2D.getExpectDataSets() - 1)) {
         svServer.mergeHandler.mergeMode2D = false;
         svServer.mergeHandler.mergeFilter2D.assignDataSet(new Long(id));
         svServer.mergeHandler.mergeFrames2D();
      } else {
         svServer.mergeHandler.mergeFilter2D.assignDataSet(new Long(id));
         monitor.display("merged ...", false);
      }
  }



  /**
   * do additional work after get the user input
   * @param command command to do work
   * @param input user input
   */
  public void doWork(int command, String input) {
    if (command == CommandSet.SAVEAS) {
       FileDialog openFileDialog = new FileDialog(this,
                         "Save DataSet", FileDialog.SAVE);
       openFileDialog.show();
       input = openFileDialog.getFile();
       if (input == null) return;

       boolean b = input.indexOf(".") >=0 ? true : false; 
       if (!b)  input += ".svb";
       input = openFileDialog.getDirectory() + input;

       writeToFile(input);

    } else if (command == CommandSet.PRINT) {
        PrintJob printjob =  getToolkit().getPrintJob(this,
                                    "SciVis 2D Plot", printerprops);
                                                        
        if (printjob != null) {          
             Graphics pg = printjob.getGraphics();
             if (pg != null) {
                  canvas.printAll(pg);
                  pg.dispose();         // flush page
             }
             printjob.end();
        }
    
    } else if (command == CommandSet.GIFSEQUENCE) {
        FileDialog openFileDialog = new FileDialog(this,
                      "GIF Sequences save", FileDialog.SAVE);
        openFileDialog.show();

        String fname = openFileDialog.getFile();
        if (fname == null) return;
        fname = openFileDialog.getDirectory() + fname;

        if (animator != null) {
           monitor.display("Animation stoped ...", false);
           animator.stop();
           animator = null;
        }

        if (gifer != null) gifer.stop();

        gifer = new GIFer(this, fname);
        gifer.start();
    } else if (command == CommandSet.CUT) {
        doCutCopy(true);
    } else if (command == CommandSet.COPY) {
        doCutCopy(false);
    } else if (command == CommandSet.PASTE) {
        Vector clipboardVec = svServer.clipboard.get2DClipboard();

        int sz = clipboardVec.size();
        if (sz > 0) {
           for (int i=0; i<sz; i++) {
              DataSet2D data = (DataSet2D)clipboardVec.elementAt(i);
              addDataSet(data);
              if (!stealthMode)
                 svServer.cmanager.broadcastDataSet(data, id);
           }
        }

    } else if (command == CommandSet.GOTO) {
        String result = JOptionPane.showInputDialog(this, 
                             "Enter the time data index to go: ");
        if (result == null) return;
        String[] args = new String[1];
        args[0] = result.trim();
        svServer.fmanager.action2d(this, CommandSet.GOTO, args,true);

    } else if (command == CommandSet2D.COLORMAP) {
        ColorMapDialog colorDialog = new ColorMapDialog(this);
        colorDialog.show();

    } else if (command == CommandSet2D.CONTOURSETTING) {
        if (canvas.contour) {
          ContourDialog contourDialog = 
             new ContourDialog(this, contour_lines, contour_min, contour_max);
          contourDialog.show();   
        } else 
          monitor.display("Not a contour mode ...", true);

    } else if (command == CommandSet.RENAMETITLE) {

        RenameDialog chooser = new RenameDialog(this);
        chooser.show();
        String[] args = chooser.getArguments();
        if (args == null) return;
        svserver.fmanager.action2d(this, CommandSet.RENAMETITLE, args, true);

    } else if (command == CommandSet.STEALTH) {
        stealthMode = stealthMI.getState();
        if (stealthMode)
          monitor.display("Stealth mode is on.", true);
        else
          monitor.display("Stealth mode is off.", true);

    } else if (command == CommandSet.LABELCHANGE) {
        String[] lbls = { config.xLabel, config.yLabel,  config.zLabel};
        String[] titles = { "X Label: " , "Y Label: ", "Z Label: " };

        LabelInputDialog chooser = 
                  new LabelInputDialog(this, titles, lbls);

        chooser.show();
        String args[] = chooser.getLabels();
        if (args != null) 
           svserver.fmanager.action2d(this, CommandSet.LABELCHANGE, args, true);
    } else if (command == CommandSet.BUFFERSIZE) {
        String bsz = JOptionPane.showInputDialog(this, 
                             "Enter the maximum buffer size: ");
        if (bsz == null) return;
        String[] args = new String[1];
        args[0] = bsz.trim();
        svServer.fmanager.action2d(this, CommandSet.BUFFERSIZE, args,true);

    } else if (command == CommandSet2D.ASPEEDCHANGE) {
        SlideInputDialog chooser = 
               new SlideInputDialog(this, "Animation Speed", 
              "Choose animation speed(msec)", 
              "Range will vary from 0 to 700 msec", 
              0, 700, animationSpeed);
        chooser.show();
        String args[] = new String[1];
        args[0] = String.valueOf(chooser.getChoosedValue());
        svserver.fmanager.action2d(this, CommandSet2D.ASPEEDCHANGE, args, true);
    } 
  }

  /**
   * animation speed change action
   */
  public void doAnimationSpeedChange(int speed) {
     if (speed < 0) animationSpeed = 0;
     else if (speed > 700) animationSpeed = 700;
     else animationSpeed = speed;

     if (animator == null) return;
     if (animator.isAlive()) {
        animator.stop();
        animator = null;
        animator = new Animator(this, animationSpeed);
        animator.start();
     }
  }


  // Cut and copy command
  private void doCutCopy(boolean cutting) {
     Vector clipboardVec = new Vector();

     if (dataSetVector.size() == 1) {
        DataSet2D data = (DataSet2D)dataSetVector.firstElement();
        
        Vector tmpvec = new Vector();
        for (int i=0; i<data.timeData.length; i++) {
           TimeData2D td;
           if (data.timeData[i].isEquallySpaced())
             td = new TimeData2D(data.timeData[i].getTime(),
              data.timeData[i].getxmin(), data.timeData[i].getxmax(),
              data.timeData[i].getDimX(),
              data.timeData[i].getymin(), data.timeData[i].getymax(),
              data.timeData[i].getDimY(), data.timeData[i].getzArr());
           else
             td = new TimeData2D(data.timeData[i].getTime(),
              data.timeData[i].getnoOfPoints(), 
              data.timeData[i].getxArr(),
              data.timeData[i].getyArr(),
              data.timeData[i].getzArr()); 

            tmpvec.addElement(td);
        }

        DataSet2D ndata = new DataSet2D(data.getTitle(), tmpvec);
        tmpvec.removeAllElements();
        tmpvec = null;

        clipboardVec.addElement(ndata);

        if (clipboardVec.size() > 0) 
           svServer.clipboard.set2DClipboard(clipboardVec);

        if (cutting) {
           removeDataSet(data);    
           svServer.cmanager.broadcastRemoveDataSet(id, data);
        }

     } else {
       if (editManager == null) 
         editManager = new CutCopyManager(this, 
                                  "Cut & Copy Selection", cutting, 2);
       editManager.show();
     }
  }

  /**
   * do multiple cut and copy
   */
  public void multiCutCopy(int[] selected, boolean cutting) {
     String str = null;
     DataSet2D data;
     Vector clipboardVec = new Vector();

     if (selected.length == 1) {
        data = (DataSet2D)(dataSetVector.elementAt(selected[0]));
        clipboardVec.addElement(data);
     } else if (selected.length > 1) {
        for (int i=0; i<selected.length; i++) {
           data = (DataSet2D)(dataSetVector.elementAt(selected[i]));
           clipboardVec.addElement(data);
        }
     } 
         
     if (clipboardVec.size() > 0) 
        svServer.clipboard.set2DClipboard(clipboardVec);

     if (selected.length == 1) {
        data = (DataSet2D)(dataSetVector.elementAt(selected[0]));
        if (cutting) {
           removeDataSet(data);       
           svServer.cmanager.broadcastRemoveDataSet(id, data);
        }

     } else if (selected.length > 1) {
        for (int i=selected.length-1; i>=0; i--) {
           data = (DataSet2D)(dataSetVector.elementAt(selected[i]));
           if (cutting) {
              removeDataSet(data);  
              svServer.cmanager.broadcastRemoveDataSet(id, data);
           }
        }
     } 
  }


  /** 
   * get next DataSet (for Animation)
   * @param done if true, repaint canvas
   */
  public void next(boolean done) {

    currentTime++;
    if (currentTime == noOfTime)
       currentTime = 0;

    if (done) {
       canvas.destroyImage();
       canvas.setDataAvailability(true);
       canvas.repaint(); 
    }
  }
  
  /** 
   * get Previous DataSet (for Animation)
   */
  public void previous() {

    currentTime--;
    if (currentTime == -1)
       currentTime = noOfTime-1;

    canvas.destroyImage();
    canvas.setDataAvailability(true);
    canvas.repaint(); 
  }


  /**
   * do animation command
   */
  public void doAnimate() {
     if (animator == null) {
        animator = new Animator(this, animationSpeed);
        animator.start();
     } else 
        System.out.println("Animation thread was corrupted ...");

     animateMI.setEnabled(false);
     nextMI.setEnabled(false);
     prevMI.setEnabled(false);
     stopMI.setEnabled(true);
  }


  /**
   * do animation  stop command
   */
  public void doStop() {
     if (animator != null) {
       animator.stop();
       animator = null;
     }
     animateMI.setEnabled(true);
     stopMI.setEnabled(false);
     nextMI.setEnabled(true);
     prevMI.setEnabled(true);
  }


  /**
   * action for boolean arguments
   */
  public void doAction_1(int id, String arg) {
     boolean b;
     if (arg.equals("true"))
         b = true;
     else
         b = false;

     switch (id) {
      case CommandSet2D.BOUNDBOX:
          config.bounding_box_2d = b;
          boundMI.setState(b);
          break;

      case CommandSet2D.XYMESH:
          config.xymesh_2d = b;
          meshMI.setState(b);
          break;

      case CommandSet2D.FACEGRID:
          config.display_grid_2d = b;
          faceMI.setState(b);
          break;

      case CommandSet2D.VECTOR:
          config.vector_plot_2d = b;
          vectorMI.setState(b);
          break;

      case CommandSet2D.XYTICK:
          config.xyticks_2d = b;
          xytickMI.setState(b);
          break;

      case CommandSet2D.ZTICK:
          config.zticks_2d = b;
          ztickMI.setState(b);
          break;

      case CommandSet.GMINMAX:
          config.global_mode = b;
          globalMI.setState(b);
          globalFlag = globalMI.getState();
          calcVertex();
          break;

      case CommandSet2D.REGENERATE:
          config.delay_regenerating_2d = b;
          delayMI.setState(b);
          return;

      case CommandSet2D.FILTEROVERLAY:

          overlayMI.setState(b);
          config.overlay_mode = overlayMI.getState();
 
          if (config.overlay_mode) {
             monitor.display("Filter overlay mode is on.", true);
             discardMI.setState(false);
          } else {
             monitor.display("Filter overlay mode is off.", true);
             discardMI.setState(true);
          }
          config.discard_mode = discardMI.getState(); 
          return;

      case CommandSet2D.FILTERDISCARD:
          discardMI.setState(b);
          config.discard_mode = discardMI.getState();
 
          if (config.discard_mode) {
             monitor.display("Filter discard mode is on.", true);
             overlayMI.setState(false);
          } else {
             monitor.display("Filter discard mode is off.", true);
             overlayMI.setState(true);
          }
          config.overlay_mode = overlayMI.getState();
          return;

      case CommandSet.DISPLAYTIME:
          timeMI.setState(b);
          config.display_time = timeMI.getState();
          break;

      case CommandSet2D.XSCALE:
          xscaleMI.setState(b);
          yscaleMI.setState(!b);
          zscaleMI.setState(!b);
          allscaleMI.setState(!b);
          zmethod = 1;
          break;

      case CommandSet2D.YSCALE:
          xscaleMI.setState(!b);
          yscaleMI.setState(b);
          zscaleMI.setState(!b);
          allscaleMI.setState(!b);
          zmethod = 2;
          break;

      case CommandSet2D.ZSCALE:
          xscaleMI.setState(!b);
          yscaleMI.setState(!b);
          zscaleMI.setState(b);
          allscaleMI.setState(!b);
          zmethod = 3;
          break;

      case CommandSet2D.ALLSCALE:
          xscaleMI.setState(!b);
          yscaleMI.setState(!b);
          zscaleMI.setState(!b);
          allscaleMI.setState(b);
          zmethod = 0;
          break;

     }

     canvas.destroyImage();
     canvas.repaint();
  }

  /**
   * action for no arguments
   */
  public void doAction_2(int id) {
    switch (id) {
      case CommandSet2D.WIRE:
          config.plotmode_2d = WIREMODE;
          if (!equallyspaced) {
              monitor.display("Unequally spaced data.", true);
              wireMI.setState(false);
              pointMI.setState(true);
              return;
          }
          wireMI.setState(true);
          cwireMI.setState(false);
          hiddenMI.setState(false);
          spectrumMI.setState(false);
          pointMI.setState(false);
          grayMI.setState(false);
          break;

      case CommandSet2D.WIRECOLOR:
          config.plotmode_2d = WIRECOLORMODE;
          if (!equallyspaced) {
              monitor.display("Unequally spaced data.", true);
              cwireMI.setState(false);
              pointMI.setState(true);
              return;
          }
          wireMI.setState(false);
          cwireMI.setState(true);
          hiddenMI.setState(false);
          spectrumMI.setState(false);
          pointMI.setState(false);
          grayMI.setState(false);
          break;

      case CommandSet2D.POINTPLOT:
          config.plotmode_2d = POINTMODE;
          wireMI.setState(false);
          cwireMI.setState(false);
          hiddenMI.setState(false);
          grayMI.setState(false);
          pointMI.setState(true);
          spectrumMI.setState(false);
          break;

      case CommandSet2D.HIDDEN:
          config.plotmode_2d = HIDDENMODE;
          if (!equallyspaced) {
              monitor.display("Unequally spaced data.", true);
              hiddenMI.setState(false);
              pointMI.setState(true);
              return;
          }
          wireMI.setState(false);
          cwireMI.setState(false);
          hiddenMI.setState(true);
          spectrumMI.setState(false);
          grayMI.setState(false);
          pointMI.setState(false);
          break;

      case CommandSet2D.COLOR:
          config.plotmode_2d = SPECTRUMMODE;
          if (!equallyspaced) {
              monitor.display("Unequally spaced data.", true);
              spectrumMI.setState(false);
              pointMI.setState(true);
              return;
          }
          wireMI.setState(false);
          cwireMI.setState(false);
          hiddenMI.setState(false);
          spectrumMI.setState(true);
          grayMI.setState(false);
          pointMI.setState(false);
          break;

      case CommandSet2D.GRAY:
          config.plotmode_2d = GRAYMODE;
          if (!equallyspaced) {
              monitor.display("Unequally spaced data.", true);
              grayMI.setState(false);
              pointMI.setState(true);
              return;
          }
          wireMI.setState(false);
          cwireMI.setState(false);
          hiddenMI.setState(false);
          spectrumMI.setState(false);
          grayMI.setState(true);
          pointMI.setState(false);
          break;

      case CommandSet2D.SURFACE:
          config.plotmode_2d = SURFACEMODE;
          surfaceMI.setState(true);
          contourMI.setState(false);
          canvas.setContour(false);
          break;

      case CommandSet2D.CONTOUR:
          config.plotmode_2d = CONTOURMODE;
          if (!equallyspaced) {
              monitor.display("Unequally spaced data.", true);
              contourMI.setState(false);
              surfaceMI.setState(true);
              return;
          }
          surfaceMI.setState(false);
          contourMI.setState(true);
          canvas.setContour(true);
          break;

    }

    canvas.destroyImage();
    canvas.repaint();
  }

  /**
   * change color map info
   */
  public void changeColorMap(String[] args) {
     svServer.fmanager.action2d(this, CommandSet.COLORMAPCHANGE, args, true);
  }
  
  /**
   * change background color
   */
  public void doBackColor(String[] args) {
    int r = Integer.valueOf(args[0]).intValue();
    int g = Integer.valueOf(args[1]).intValue();
    int b = Integer.valueOf(args[2]).intValue();
    config.bgColor = new Color(r, g, b);

    canvas.destroyImage();
    canvas.repaint();
  }


  /**
   * label change action
   */
  public void doLabelChange(String args[]) {
      config.xLabel = args[0];
      config.yLabel = args[1];
      config.zLabel = args[2];
      canvas.destroyImage();
      canvas.repaint();
  }

  /**
   * split command
   */
  public void doSplit() {

     DataSet2D data = (DataSet2D)dataSetVector.lastElement();

     long frameID  = svServer.fmanager.register(data, -1, null);
     removeDataSet(data);
     canvas.repaint();
  }

  /**
   * goto action
   */
  public void doGotoAction(String input) {
      int whereto = Integer.valueOf(input.trim()).intValue();

      if ((whereto > noOfTime) || (whereto <= 0))
         monitor.display("Invalid range ...", true);

      int index = 0;
      while (whereto != currentTime) {
         next(false);

         if (index > noOfTime) break;
         index++;
       }
      previous();
  }


  /**
   * buffer size change action
   */
  public void doBufferAction(String input) {
      int size = Integer.valueOf(input.trim()).intValue();

      if (size <= 0)
         monitor.display("Invalid range ...", true);

      bufferSize = size;

      for (Enumeration e = dataSetVector.elements(); e.hasMoreElements(); ) {
        DataSet2D data = (DataSet2D)e.nextElement();
        data.setBufferSize(size);
      }
  }


  /**
   * update dataSet to this frame
   * @param pos position of DataSet to be replaced
   * @param dataSet DataSet to be updated
   */
  public synchronized void updateDataSet(int pos, DataSet2D dataSet) {
     dataSetVector.removeElementAt(pos);
     dataSetVector.addElement(dataSet);
     updateTimeArr();

     calcGlobalMinMax();
     calcVertex();

     if (currentTime != noOfTime-1) {
        int saved = currentTime;
        currentTime = 0;

        // send to last time
        for (int i=0; i<saved; i++)
           next(false);
        next(true);   // one more time
     }

     canvas.makeContourCache();
     monitor.display("Updated DataSet " + dataSet.getTitle(), false);
  }


  /**
   * update dataSets to this frame
   */
  public void updateDataSet() {

     calcGlobalMinMax();
     calcVertex();

     canvas.makeContourCache();
     canvas.destroyImage();
     canvas.repaint();
  }

  /**
   * remove dataSet from this frame
   * @param str DataSet title to be removed
   */
  public void removeDataSet(String str) {
     
    for (Enumeration e=dataSetVector.elements(); e.hasMoreElements();) {
       DataSet2D data = (DataSet2D)e.nextElement();

       if (str.equals(data.getTitle())) {
          removeDataSet(data);
          break;
       }
    } 
    canvas.makeContourCache(); 
    canvas.repaint(); 
  }

  /**
   * remove a dataSet from this frame
   * @param dataSet DataSet to be removed
   */
  public void removeDataSet(DataSet2D dataSet) {
     dataSetVector.removeElement(dataSet);

     // reset all the currentTime value to 0
     currentTime = 0;

     if (dataSetVector.size() <= 0) {
        svServer.fmanager.removeNotify(id, DataSet.TWOD);
        this.dispose();
     } else {
         DataSet2D data = (DataSet2D)dataSetVector.firstElement();
         if (dataSetVector.size() == 1) {
            setFrameTitle(data.getTitle());
            splitMI.setEnabled(false);
         } else
            setFrameTitle(data.getTitle() + "...");
     }

     calcVertex();
     updateTimeArr();
     monitor.display("Removed DataSet " + dataSet.getTitle(), false);
     canvas.makeContourCache();
     canvas.repaint();
  }



  /**
   * kill this frame
   */
  public void killFrame() {

      if (animator != null) {
        animator.stop();
        animator = null;
      } 
      if (gifer != null) {
        gifer.stop();
        gifer = null;
      } 

      dataSetVector.removeAllElements();
      dataSetVector = null;

      canvas = null;
      svServer.fmanager.removeNotify(id, DataSet.TWOD);
      this.dispose();
   }


  // write all objects of this frame to a file
  public void writeToFile(String fname) {
     DataOutputStream os = null;
     byte[] byteArr;
     int sz = dataSetVector.size();
   
     try {
        os = new DataOutputStream(
               new BufferedOutputStream(new FileOutputStream(fname)));
     } catch (Exception e) {
        System.out.println("File writing error(2d) ...");
        return;
     }

     try {
        // write dimension
        os.writeInt(DataSet.TWOD);    

        // write frame title
        byteArr = frameTitle.getBytes();
        os.writeInt(byteArr.length); 
        os.write(byteArr, 0, byteArr.length);
              
        // write configuration stream
        byteArr = config.getFileSettings().getBytes();
        os.writeInt(byteArr.length); 
        os.write(byteArr, 0, byteArr.length);

        // write drawing objects stream (NOT YET IMPLEMENTED)
        os.writeInt(0); 

        // write the number of data sets
        os.writeInt(sz);  

        // now write the data sets
        for (int i=0; i<sz; i++) {
           DataSet2D dataSet = (DataSet2D)dataSetVector.elementAt(i);

           byteArr = dataSet.getTitle().getBytes();
           os.writeInt(byteArr.length); 
           os.write(byteArr, 0, byteArr.length);

           int noOfTimeData = dataSet.timeData.length;
           os.writeInt(noOfTimeData);   // send number of TimeData
           for (int k=0; k<noOfTimeData; k++) 
              dataSet.timeData[k].sendingData(os);
        } 

        os.close();
     } catch (Exception e) {
        System.out.println("Can't write the 2d data to a file ...");
     }     
  }


   // make menu system
   private void makeMenu() {
     MenuBar menubar = new MenuBar();
     CommandSet2D command;
     MenuItem mi;
     int filter_command = 1000;

     // File manager
     Menu fileMenu = new Menu("File");

     addMI("Save as ...", fileMenu, new CommandSet2D(this, CommandSet.SAVEAS));
     addMI("Print", fileMenu, new CommandSet2D(this, CommandSet.PRINT));
     addMI("GIF Sequences", fileMenu, 
                              new CommandSet2D(this, CommandSet.GIFSEQUENCE));
     fileMenu.addSeparator();
     addMI("Quit C-q", fileMenu, new CommandSet2D(this, CommandSet.QUIT));

     menubar.add(fileMenu);

     // Cut, Copy and Paste
     Menu editMenu = new Menu("Edit");
     addMI("Cut", editMenu, new CommandSet2D(this, CommandSet.CUT));
     addMI("Copy", editMenu, new CommandSet2D(this, CommandSet.COPY));
     addMI("Paste", editMenu, new CommandSet2D(this, CommandSet.PASTE));

     menubar.add(editMenu);

     // Configuration
     Menu configMenu = new Menu("View");

     boundMI = new CheckboxMenuItem("Bound box");
     command = new CommandSet2D(this, CommandSet2D.BOUNDBOX);
     boundMI.addItemListener(command);
     boundMI.setState(config.bounding_box_2d); 
     configMenu.add(boundMI);

     meshMI = new CheckboxMenuItem("X-Y Mesh");
     command = new CommandSet2D(this, CommandSet2D.XYMESH);
     meshMI.addItemListener(command);
     meshMI.setState(config.xymesh_2d); 
     configMenu.add(meshMI);


     faceMI = new CheckboxMenuItem("Face Grid");
     command = new CommandSet2D(this, CommandSet2D.FACEGRID);
     faceMI.addItemListener(command);
     faceMI.setState(config.display_grid_2d); 
     configMenu.add(faceMI);

     vectorMI = new CheckboxMenuItem("Vector Plot");
     command = new CommandSet2D(this, CommandSet2D.VECTOR);
     vectorMI.addItemListener(command);
     vectorMI.setState(config.vector_plot_2d); 
vectorMI.setEnabled(false);
     configMenu.add(vectorMI);

     configMenu.addSeparator();

     wireMI = new CheckboxMenuItem("Wire Frame");
     command = new CommandSet2D(this, CommandSet2D.WIRE);
     wireMI.addItemListener(command);
     if (config.plotmode_2d == WIREMODE)
        wireMI.setState(true); 
     else
        wireMI.setState(false); 
     configMenu.add(wireMI);

     cwireMI = new CheckboxMenuItem("Colored Wire Frame");
     command = new CommandSet2D(this, CommandSet2D.WIRECOLOR);
     cwireMI.addItemListener(command);
     if (config.plotmode_2d == WIRECOLORMODE)
        cwireMI.setState(true); 
     else
        cwireMI.setState(false); 
     configMenu.add(cwireMI);

     hiddenMI = new CheckboxMenuItem("Hidden Surface Eliminated");
     command = new CommandSet2D(this, CommandSet2D.HIDDEN);
     hiddenMI.addItemListener(command);
     if (config.plotmode_2d == HIDDENMODE)
        hiddenMI.setState(true); 
     else
        hiddenMI.setState(false); 
     configMenu.add(hiddenMI);

     spectrumMI = new CheckboxMenuItem("Color Mapped");
     command = new CommandSet2D(this, CommandSet2D.COLOR);
     spectrumMI.addItemListener(command);
     if (config.plotmode_2d == SPECTRUMMODE)
        spectrumMI.setState(true); 
     else
        spectrumMI.setState(false); 
     configMenu.add(spectrumMI);

     grayMI = new CheckboxMenuItem("Gray Scale");
     command = new CommandSet2D(this, CommandSet2D.GRAY);
     grayMI.addItemListener(command);
     if (config.plotmode_2d == GRAYMODE)
        grayMI.setState(true); 
     else
        grayMI.setState(false); 
     configMenu.add(grayMI);

     pointMI = new CheckboxMenuItem("Point Plot");
     command = new CommandSet2D(this, CommandSet2D.POINTPLOT);
     pointMI.addItemListener(command);
     if (config.plotmode_2d == POINTMODE ||  config.plotmode_2d == SURFACEMODE)
        pointMI.setState(true); 
     else
        pointMI.setState(false); 
     configMenu.add(pointMI);

     configMenu.addSeparator();

     surfaceMI = new CheckboxMenuItem("Surface Plot");
     command = new CommandSet2D(this, CommandSet2D.SURFACE);
     surfaceMI.addItemListener(command);
     if (config.plotmode_2d != 6)
        surfaceMI.setState(true); 
     else
        surfaceMI.setState(false); 
     configMenu.add(surfaceMI);

     contourMI = new CheckboxMenuItem("Contour");
     command = new CommandSet2D(this, CommandSet2D.CONTOUR);
     contourMI.addItemListener(command);
      if (config.plotmode_2d == CONTOURMODE)
        contourMI.setState(true); 
     else
        contourMI.setState(false); 
     configMenu.add(contourMI);

     configMenu.addSeparator();


     Menu colorMenu = new Menu("Colors");

     mi = new MenuItem("Background Color");
     ActionListener bgColorChooser = new ActionListener() {
	public void actionPerformed(ActionEvent e) {
           Color c = 
              JColorChooser.showDialog(GFrame2D.this, 
                                      "Background Color", config.bgColor);
           if (c == null)
               return;
           String args[] = { String.valueOf(c.getRed()), 
                             String.valueOf(c.getGreen()), 
                             String.valueOf(c.getBlue()) };
           svServer.fmanager.action2d(GFrame2D.this, 
                                    CommandSet.BACKCOLOR, args, true);
	}
     };
     mi.addActionListener(bgColorChooser);
     colorMenu.add(mi);

     mi = new MenuItem("Color Map Editor");
     command = new CommandSet2D(this, CommandSet2D.COLORMAP);
     mi.addActionListener(command);
     colorMenu.add(mi);

     configMenu.add(colorMenu);

     delayMI = new CheckboxMenuItem("Delay Regenerating");
     command = new CommandSet2D(this, CommandSet2D.REGENERATE);
     delayMI.addItemListener(command);
     delayMI.setState(config.delay_regenerating_2d); 
     configMenu.add(delayMI);

     timeMI = new CheckboxMenuItem("Display Time");
     command = new CommandSet2D(this, CommandSet.DISPLAYTIME);
     timeMI.addItemListener(command);
     timeMI.setState(config.display_time);
     configMenu.add(timeMI);

     Menu ticksMenu = new Menu("Label & Ticks");

     addMI("Change Labels", ticksMenu, 
                         new CommandSet2D(this, CommandSet.LABELCHANGE));

     xytickMI = new CheckboxMenuItem("X-Y ticks");
     command = new CommandSet2D(this, CommandSet2D.XYTICK);
     xytickMI.addItemListener(command);
     xytickMI.setState(config.xyticks_2d); 
     ticksMenu.add(xytickMI);

     ztickMI = new CheckboxMenuItem("Z ticks");
     command = new CommandSet2D(this, CommandSet2D.ZTICK);
     ztickMI.addItemListener(command);
     ztickMI.setState(config.zticks_2d); 
     ticksMenu.add(ztickMI);

     configMenu.add(ticksMenu);

     addMI("Rename Title", configMenu, 
                           new CommandSet2D(this, CommandSet.RENAMETITLE));

     globalMI = new CheckboxMenuItem("Use global minmax");
     command = new CommandSet2D(this, CommandSet.GMINMAX);
     globalMI.addItemListener(command);
     globalMI.setState(config.global_mode);
     configMenu.add(globalMI);

     menubar.add(configMenu);


     ////////////////////////////////////////////////////
     // FILTERS
     ////////////////////////////////////////////////////
     filterMenu = new Menu("Filter");

     createFilterMenu();
     menubar.add(filterMenu);


     // Misc menu
     Menu miscMenu = new Menu("Misc");

     // animation menus
     Menu animationMenu = new Menu("Animation");

     animateMI = new MenuItem("Animate C-a");
     command = new CommandSet2D(this, CommandSet.ANIMATE);
     animateMI.addActionListener(command);
     animationMenu.add(animateMI);

     stopMI = new MenuItem("Stop C-s");
     command = new CommandSet2D(this, CommandSet.STOP);
     stopMI.addActionListener(command);
     animationMenu.add(stopMI);
     stopMI.setEnabled(false);

     nextMI = new MenuItem("Next C-n");
     command = new CommandSet2D(this, CommandSet.NEXT);
     nextMI.addActionListener(command);
     animationMenu.add(nextMI);

     prevMI = new MenuItem("Previous C-p");
     command = new CommandSet2D(this, CommandSet.PREVIOUS);
     prevMI.addActionListener(command);
     animationMenu.add(prevMI);

     addMI("Animation Speed", animationMenu, 
                    new CommandSet2D(this, CommandSet2D.ASPEEDCHANGE));

     miscMenu.add(animationMenu);

     // scale menus
     Menu scaleMenu = new Menu("Scale");

     xscaleMI = new CheckboxMenuItem("X Scale");
     command = new CommandSet2D(this, CommandSet2D.XSCALE);
     xscaleMI.addItemListener(command);
     scaleMenu.add(xscaleMI);

     yscaleMI = new CheckboxMenuItem("Y Scale");
     command = new CommandSet2D(this, CommandSet2D.YSCALE);
     yscaleMI.addItemListener(command);
     scaleMenu.add(yscaleMI);

     zscaleMI = new CheckboxMenuItem("Z Scale");
     command = new CommandSet2D(this, CommandSet2D.ZSCALE);
     zscaleMI.addItemListener(command);
     scaleMenu.add(zscaleMI);

     allscaleMI = new CheckboxMenuItem("XYZ Scale");
     command = new CommandSet2D(this, CommandSet2D.ALLSCALE);
     allscaleMI.addItemListener(command);
     scaleMenu.add(allscaleMI);
     allscaleMI.setState(true);
     miscMenu.add(scaleMenu);

     mi = new MenuItem("Contour Setting");
     command = new CommandSet2D(this, CommandSet2D.CONTOURSETTING);
     mi.addActionListener(command);
     miscMenu.add(mi);

     addMI("Buffer Size", miscMenu, 
                    new CommandSet2D(this, CommandSet.BUFFERSIZE));

     mi = new MenuItem("Goto C-g");
     command = new CommandSet2D(this, CommandSet.GOTO);
     mi.addActionListener(command);
     miscMenu.add(mi);

     addMI("Reset", miscMenu, new CommandSet2D(this, CommandSet2D.RESET));

     // Split the DataSet
     splitMI = new MenuItem("Split DataSets");
     command = new CommandSet2D(this, CommandSet.SPLIT);
     splitMI.addActionListener(command);
     miscMenu.add(splitMI);
     if (dataSetVector.size() < 2) 
        splitMI.setEnabled(false);

     // Stealth mode
     stealthMI = new CheckboxMenuItem("Stealth Mode");
     command = new CommandSet2D(this, CommandSet.STEALTH);
     stealthMI.addItemListener(command);
     stealthMI.setState(stealthMode);
     miscMenu.add(stealthMI);
    
     menubar.add(miscMenu);

     setMenuBar(menubar);
  }

  // add menuItem to menu
  private void addMI(String str, Menu parent, CommandSet2D command) {
     MenuItem mi = new MenuItem(str);
     mi.addActionListener(command);
     parent.add(mi);
  }



  // slide value changes
  private void createSlideListener() {
     ChangeListener changeListener = new ChangeListener() {
        public void stateChanged(ChangeEvent e) {

           zfactor = zoomer.getValue();
           float value = zfactor / 20.f;
           
           String val = String.valueOf(value);
           if (val.length() > 4)
              val = val.substring(0, 4);
           
	   zoomValue.setText(val);
	   canvas.setScale(zmethod, value);
        }
     };
     zoomer.addChangeListener(changeListener);
  }



  // virtual desktop management
  private boolean iconized = false;
  private boolean activated = true;

  public boolean isIconized() {
     return iconized;
  }

  public boolean isActivated() {
     return activated;
  }

  public void windowOpened(WindowEvent e) {
  }
  public void windowClosing(WindowEvent e) {
  }

  public void windowClosed(WindowEvent e) {
  }

  public void windowIconified(WindowEvent e) {
    iconized = true;
    if (svServer.controller != null)
       svServer.controller.desktopRepaint();
  }

  public void windowDeiconified(WindowEvent e) {
    iconized = false;
    if (svServer.controller != null)
       svServer.controller.desktopRepaint();
  }

  public void windowActivated(WindowEvent e) {
    activated = true;
    if (svServer.controller != null)
       svServer.controller.desktopRepaint();
  }

  public void windowDeactivated(WindowEvent e) {
    activated = false;
    if (svServer.controller != null)
       svServer.controller.desktopRepaint();
  }


  private void createFilterMenu() {
     CommandSet2D command;
     MenuItem mi;
     int filter_command = 1000;

     overlayMI = new CheckboxMenuItem("Overlay Mode");
     command = new CommandSet2D(this, CommandSet.FILTEROVERLAY);
     overlayMI.addItemListener(command);
     overlayMI.setState(config.overlay_mode);
     filterMenu.add(overlayMI);

     discardMI = new CheckboxMenuItem("Discard Mode");
     command = new CommandSet2D(this, CommandSet.FILTERDISCARD);
     discardMI.addItemListener(command);
     discardMI.setState(config.discard_mode);
     filterMenu.add(discardMI);

     filterMenu.addSeparator();

     Vector level0Vec = new Vector();
     Vector level1Vec = new Vector();
     Vector level2Vec = new Vector();
     Vector level3Vec = new Vector();

     for (Enumeration e=svServer.menuHandler.get2dElements(); 
                                               e.hasMoreElements();) {
        MenuInfo info = (MenuInfo)e.nextElement();

        if (info.level == 0) 
	  level0Vec.addElement(info);
        else if (info.level == 1) 
	  level1Vec.addElement(info);
        else if (info.level == 2) 
	  level2Vec.addElement(info);
        else if (info.level == 3) 
	  level3Vec.addElement(info);
     } 

     for (Enumeration e=level0Vec.elements(); e.hasMoreElements();) {
        MenuInfo lv0item = (MenuInfo)e.nextElement();

        if (!lv0item.isItem) {
           Menu lv0Menu = new Menu(lv0item.name);

           for (Enumeration e1=level1Vec.elements(); e1.hasMoreElements();) {
              MenuInfo lv1item = (MenuInfo)e1.nextElement();
              
              if (lv1item.parent == lv0item.id) {
                 if (!lv1item.isItem) {
                    Menu lv1Menu = new Menu(lv1item.name);
                    
                    for (Enumeration e2=level2Vec.elements(); 
                                               e2.hasMoreElements();) {
                       MenuInfo lv2item = (MenuInfo)e2.nextElement();

                       if (lv2item.parent == lv1item.id) {
                          if (!lv2item.isItem) {
                             Menu lv2Menu = new Menu(lv2item.name);

                             for (Enumeration e3=level3Vec.elements(); 
                                               e3.hasMoreElements();) {
                                MenuInfo lv3item = (MenuInfo)e3.nextElement();
                                if (lv3item.parent == lv2item.id) {
                                   if (lv3item.isItem)  
                                      addMI(lv3item.name, lv2Menu, new CommandSet2D(this, ++filter_command));
                                }
			     }

                             lv1Menu.add(lv2Menu);
                          } else {
                             addMI(lv2item.name, lv1Menu, new CommandSet2D(this, ++filter_command));
                          }
                       }
                    }

                    lv0Menu.add(lv1Menu);
                 } else {
                    addMI(lv1item.name, lv0Menu, new CommandSet2D(this, ++filter_command));
                 }
              }
           }
           filterMenu.add(lv0Menu);
        } else {
           addMI(lv0item.name, filterMenu, new CommandSet2D(this, ++filter_command));
        }
     }
  }

} // end of class



